 /*******************************************************************************
	ProjectName: Capacitance Meter

	FileName:    main.c

  Copyright(c) 2022.10  Suwa-Koubou

   pin assign PIC32MX220F032B

                                 MCLR 1    28 AVDD       ------ 3.3V
          debug to PC <----  U1TX/RA0 2    27 AVSS       ------ GND
                                  RA1 3    26 RB15                    
                 0.7VDD --- C2INB/RB0 4    25 RB14                            
           C -+------------ C2INA/RB1 5    24 RB13                            
              |  0.2VDD --- C1INB/RB2 6    23 Vusb3v3
              +------------ C1INA/RB3 7    22 RB11/PGEC2  
                 GND  -----   GND/VSS 8    21 RB10/PGED2 
           20MHz    -+-----  OSC1/RA2 9    20 VCAP via 10uF to  GND
           Cerarock  +----- OSC2/RA3 10    19 GND       ------- GND  
            to C ----- 1Kohm --  RB4 11    18 RB9/SDA   ------- OLED/SDA
              Act LED <----      RA4 12    17 RB8/SCL   ------- OLED/SCL
                3.3V --------    VDD 13    16 RB7                            
                Sw ----------    RB5 14    15 VBUS


      Timer1   1msec interval timer (include delay_ms() function)
      Timer2
      Timer3
      Timer4   32bit Timer PRx count 1 - 4.2G
      Timer5   32bit Timer PRy

      interrupt level
        Timer1  1
        UART1   2
        COMP1   3
        COMP2   3


 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "HardwareProfile.h"
#include "typedef.h"
#include "timer.h"

#include "debug.h"

/*------------------------------------------------------------------------------
	chip Configuration PIC32MX220F032B  without USB
------------------------------------------------------------------------------*/

#include "p32mx220f_hs20_config.h"


/*------------------------------------------------------------------------------
	definition
------------------------------------------------------------------------------*/

#define act_led_on()     LATAbits.LATA4=0
#define act_led_off()    LATAbits.LATA4=1

#define mainte_sw        PORTBbits.RB5

#define SW_ON            0
#define SW_OFF           1


// charging resistance type
#define  SEL_LOW_R     0    // 1Mohm+1Kohm pararell
#define  SEL_HIGH_R    1    // 1Mohm only

/*------------------------------------------------------------------------------
	global variables
------------------------------------------------------------------------------*/

// 32bit timer capture value
unsigned int t_020Vdd;
unsigned int t_070Vdd;
unsigned int t_CR;        // t_CR=(t_070Vdd-t_020Vdd)*2
unsigned int stray_capa;  // stray capacity of Measuring Circuit

// measure done flag
volatile int measure_done;

char pbuf[32];


/*------------------------------------------------------------------------------
	function prototype
------------------------------------------------------------------------------*/

int   main (void);
void  maintenance(void);
void  measure_capacitance(void);
void  measure_Time_Constant(int select_charge_resistor);
void  io_initialize(void);


/*==============================================================================
	main() start here. 
==============================================================================*/

int  main ( void )
{
    io_initialize();
    oled_init();

    act_led_on();

    for(;;){
        if(mainte_sw == SW_ON)maintenance();
        measure_capacitance();
        delay_ms(500);
    }

    return 0;
}



void maintenance(void)
{
    unsigned int TC;
    int          n;

    oled_clear();
    oled_putstr2(0,0,"Stray Capacity  ");
    oled_putstr2(0,2,"  measuring ... ");

    TC = 0;
    for(n=0; n<6; n++){
        measure_Time_Constant(SEL_HIGH_R);
        if(measure_done){
            TC += t_CR;
        }else{
            n--;
        }
    }
    stray_capa = TC/6;

    while(mainte_sw == SW_ON);

   oled_putstr2(0,2,"  done!         ");
}



/*=========================================================================
    measuring capacitance
=========================================================================*/

/*------------------------------------------------------------------------------
    measure_capacitance(void)
------------------------------------------------------------------------------*/

int  put_Capa(unsigned int measure_val, unsigned int unit_val, char *unit);

void measure_capacitance(void)
{
    int           sel_R;
    unsigned int  val;


    // 1st. select High Resistor
    sel_R = SEL_HIGH_R;
    measure_Time_Constant(sel_R);
    if(measure_done ==0){
        // 2nd. select Low Resistor 
        sel_R = SEL_LOW_R;
        measure_Time_Constant(sel_R);
    }

    if(measure_done ==0){
        // can't measure,  during 500msec time over
        oled_putstr4(0,2, "----");
        return;
    }


    // calculate Capacitance from t_CR(Time Constant)
    if(sel_R == SEL_HIGH_R){  // R=1Mohm

        // adjustment stray capacity, high resistor only
        if(t_CR >= stray_capa){
            t_CR -= stray_capa;
        }else{
            t_CR = 0;
        }        

        if(put_Capa(t_CR, 40000000UL, "uF") == -1){
            if(put_Capa(t_CR, 40000UL, "nF") == -1){
                if(put_Capa(t_CR, 40UL, "pF")== -1){
                    oled_putstr4(0,0, "  0p");
                }
            }
        }
    }

    if(sel_R == SEL_LOW_R){  // R=1Kohm
        if(put_Capa(t_CR, 40000UL, "uF") == -1){
            put_Capa(t_CR, 40UL, "nF");
        }
    }

}


int  put_Capa(unsigned int measure_val, unsigned int unit_val, char *unit)
{
    int val, col;
    char buf[17];


    val = measure_val / unit_val;
    if(val == 0) return -1;

    if(val < 10){   // x.xx  x.00
        buf[0] = val+'0';
        if(unit_val/100){
            val = measure_val / (unit_val/100);
            val %= 100;
            buf[1] = (val/10) +'0';
            buf[2] = (val % 10)+'0';
        }else{
            buf[1] = '0';
            buf[2] = '0';
        }
        col = 4;

    }else if(val < 100){    // xx.x   xx.0
        buf[0] = (val/10) + '0';
        buf[1] = (val % 10) + '0';
        if(unit_val/10){
            val = measure_val / (unit_val/10);
            buf[2] = (val % 10)+'0';
        }else{
            buf[2] = '0';
        }
        col = 8;

    }else{  // xxx
        buf[0] = (val/100) + '0';
        val %= 100;
        buf[1] = (val/10) + '0';
        buf[2] = (val%10) + '0';
        col = 0;
    }
    strcpy(&buf[3], unit);
    buf[4] = 0;  // delete 'F'
    oled_putstr4(0,0, buf);

    if(col){
        oled_locate(col, 3);
        oled_putch('.');
    }

    return 0;
}


/*------------------------------------------------------------------------------
    measure_Time_Constant(int select_charge_resistor)
------------------------------------------------------------------------------*/

void measure_Time_Constant(int select_charge_resistor)
{
    int  tid;

    if(select_charge_resistor == SEL_HIGH_R){
        TRISBbits.TRISB4=1;  // dis connect 1Kohm. charge by 1Mohm only.
    }else{
        TRISBbits.TRISB4=0;  // connect 1Kohm. charge by 1Mohm//1Kohm .
        LATBbits.LATB4=1;
    }

    // Capacitance discharge
    TRISBbits.TRISB3=0;  // to GND C1INA(+) input
    LATBbits.LATB3=0;    
    delay_ms(500);       // discharge during 500msec

    // go measuring
    t_CR = 0;            // clear CR(Time Constant)
    measure_done = 0;    // reset complete flag

    tid = timer_entry(500);  // done wait 500msec

    IFS1bits.CMP1IF=0;   // clear interrupt flag
    IFS1bits.CMP2IF=0;   // clear interrupt flag
    IEC1bits.CMP1IE=1;   // enable comp1 interrupt
    IEC1bits.CMP2IE=1;   // enable comp2 interrupt
    TRISBbits.TRISB3=1;  // release C1INA(+), change to INPUT mode
    CM1CONbits.ON=1;     // activate comp1 module
    CM2CONbits.ON=1;     // activate comp2 module

    while(measure_done ==0 && timer_check(tid));

    CM1CONbits.ON=0;     // none activate comp1 module
    CM2CONbits.ON=0;     // none activate comp2 module
    IEC1bits.CMP1IE=0;   // disable comp1 interrupt
    IEC1bits.CMP2IE=0;   // disable comp2 interrupt
    IFS1bits.CMP1IF=0;
    IFS1bits.CMP2IF=0;

    timer_deletex(tid);

    return;
}

/*==============================================================================
    interrupt functions of Comparator module
==============================================================================*/
//
//  0.2VDD comparator transition interrupt
//
void __attribute__ (( interrupt(ipl3soft), vector(_COMPARATOR_1_VECTOR) )) _CMP1Interrupt(void)
{
    t_020Vdd = TMR4;

    IFS1bits.CMP1IF = 0;
    return;
}

//
//  0.7VDD comparator transition interrupt
//
void __attribute__ (( interrupt(ipl3soft), vector(_COMPARATOR_2_VECTOR) )) _CMP2Interrupt(void)
{
    t_070Vdd = TMR4;
    t_CR = ((t_070Vdd - t_020Vdd)<<1);
    measure_done = 1;

    IFS1bits.CMP2IF = 0;
    return;
}


/*==============================================================================
	io initialize
==============================================================================*/

void io_initialize(void)
{
    __builtin_disable_interrupts();
    
    // Data Memory SRAM wait states: Default Setting = 1; set it to 0
    BMXCONbits.BMXWSDRM = 0;

    // enable multi-vector interrupts
    INTCONSET = _INTCON_MVEC_MASK;
 
    // enable global interrupts
    __builtin_enable_interrupts();


    //--- setting I/O port ----------------------------------------
    ANSELA=0;
    ANSELB=0;

    LATA = 0x0000; TRISA = 0xffff;	// RA are default input.
    LATB = 0x0000; TRISB = 0xffff;	// RB are default input.

    // *** Power LED **************************************
    LATAbits.LATA4=1;      // act LED turn off
    TRISAbits.TRISA4=0;

    // *** UART1 ******************************************
    RPA0R = 0b0001;         // RA0 is assigned to U1TX
    LATAbits.LATA0 = 1;		// RA0(U1TX) to level high
    TRISAbits.TRISA0 = 0;	// RA0(U1TX) is output
    // can't use U1RX
    //U1RXR = 0b0100;         // RB2 is assigned to U1RX
    //TRISBbits.TRISB2=1;	    // RB2 is input 

    // *** stray capacity measure switch ***************
    TRISBbits.TRISB5=1;
    CNPUBbits.CNPUB5=1;    // weak pullup

    // *** set capacitance volatge comparator
    TRISBbits.TRISB0 = 1;  // C2INB(-) 0.7VDD
    TRISBbits.TRISB1 = 1;  // C2INA(+) Capa monitor reached 0.7VDD
    TRISBbits.TRISB2 = 1;  // C1INB(-) 0.2VDD
    TRISBbits.TRISB3 = 1;  // C1INA(+) Capa monitor reached 0.2VDD

    // *** capacitance charg and discharg
    TRISBbits.TRISB4 = 0; // charge by 1kohm 
    LATBbits.LATB4 = 1;   // VDD out

    // *** set comparator module
    CM1CON=0;    // CCH(-)=0b00 select INB, CREF(+)=0b0 select INA
    CM1CONbits.EVPOL=0b01;  // interrupt low to high
    IPC6bits.CMP1IP=3;
    IFS1bits.CMP1IF=0;
    //IEC1bits.CMP1IE=1;
    //CM1CONbits.ON=1;  // go comparate

    CM2CON=0;    // CCH(-)=0b00 select INB, CREF(+)=0b0 select INA
    CM2CONbits.EVPOL=0b01;  // interrupt low to high
    IPC7bits.CMP2IP=3;
    IFS1bits.CMP2IF=0;
    //IEC1bits.CMP2IE=1; 
    //CM2CONbits.ON=1;  // go comparate


    // setup Timer4 and Timer5 32bit counter
    T4CON = 0;
    T5CON = 0;
    T4CONbits.T32 = 1;         // 32bit T4+T5 
    T4CONbits.TCS = 0;         // clock in from peripheral: clock is 40MHz/TCKPS
    T4CONbits.TCKPS = 0;       // prescale 1:1  40MHz/1=40MHz
    IEC0bits.T4IE = 0;         // don't use interrupt
    IEC0bits.T5IE = 0;
    PR4 = 0xffffffff;          // load PR4, PR5 with 32bit
    T4CONbits.TON=1;           // Go 40MHz clock count 


    // *** interval timer (use Timer1) *******************
    timer_init();

    // *** Init UART *************************************
    UART1Init();

    // *** OLED  I2C interface enable *********************
    I2C1CON = 0;
    I2C1BRG = 49;  // PBCLK=40MHz(40000KHz), FSCK=400KHz,
                   // (PBCLK/(2*FSCK))-1=(40000/(2*400))-1=49
    I2C1CONbits.DISSLW = 1;
    I2C1CONbits.I2CEN = 1;

}

/*** end of main.c ************************************************************/

