/*

UART1 Driver File for PIC24.

********************************************************************************
 FileName:        UART1.c
 Dependencies:    HardwareProfile.h
 Processor:       PIC24
 Compiler:        MPLAB C30
 Linker:          MPLAB LINK30
 Company:         Microchip Technology Incorporated

Author                Date      Comment
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
KO                 12-Feb-2008  Modified to use HardwareProfile.h
KO                 11-Oct-2006  v1.0
Anton Alkhimenok   18-Oct-2005
Anton Alkhimenok   17-Feb-2009  Added UART1Char2Hex(), UART1Hex2Char(), 
									  UART1ClearError(), UART1DataReceived()
PAT				   27-Jan-2010  Added UART1GetBaudError() for dynamic checking
								of baud rate percentage error.

********************************************************************************
Software License Agreement

Microchip Technology Inc. ("Microchip") licenses to you the right to use, copy,
modify and distribute the software - including source code - only for use with
Microchip microcontrollers or Microchip digital signal controllers; provided
that no open source or free software is incorporated into the Source Code
without Microchips prior written consent in each instance.

The software is owned by Microchip and its licensors, and is protected under
applicable copyright laws.  All rights reserved.

SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, ANY CLAIMS BY THIRD PARTIES (INCLUDING
BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

********************************************************************************
*/

//#include "Compiler.h"
#include "HardwareProfile.h"
#include "uart1.h"

//******************************************************************************
// Constants
//******************************************************************************

//U1BRG register value and baudrate mistake calculation

#if defined (__C30__) || defined ( __XC16__ )
    #if defined (__dsPIC33E__) || defined (__PIC24E__)
    #define BAUDRATEREG1        (((GetPeripheralClock())/(BRG_DIV1 * BAUDRATE1)) - 1L)
    #else
    #define BAUDRATEREG1        (((GetSystemClock()/2L)+(BRG_DIV1/2L*BAUDRATE1))/BRG_DIV1/BAUDRATE1-1L)
	#endif
#elif defined (__PIC32MX__)
    #define BAUDRATEREG1        ((GetPeripheralClock()+(BRG_DIV1/2L*BAUDRATE1))/BRG_DIV1/BAUDRATE1-1L)
#else
    #error Cannot calculate BRG value
#endif    

#if defined (__C30__) || defined ( __XC16__ )
    #if defined (__dsPIC33E__)|| defined (__PIC24E__)
    #define BAUD_ACTUAL         ((GetPeripheralClock())/(BRG_DIV1 * (BAUDRATEREG1+1L)))
    #else
    #define BAUD_ACTUAL         ((GetSystemClock()/2L)/BRG_DIV1/(BAUDRATEREG1+1L))
	#endif
#elif defined (__PIC32MX__)
    #define BAUD_ACTUAL         (GetPeripheralClock()/BRG_DIV1/(BAUDRATEREG1+1L))
#else
    #error Cannot calculate actual baud rate
#endif    

	#define BAUD_ERROR              ((BAUD_ACTUAL > BAUDRATE1) ? BAUD_ACTUAL-BAUDRATE1 : BAUDRATE1-BAUD_ACTUAL)
	#define BAUD_ERROR_PERCENT      ((BAUD_ERROR*100L+BAUDRATE1/2L)/BAUDRATE1)
	
#if defined (__C30__) || defined ( __XC16__ )

	#if (BAUD_ERROR_PERCENT > 3)
	    #error UART frequency error is worse than 3%
	#elif (BAUD_ERROR_PERCENT > 2)
	    #warning UART frequency error is worse than 2%
	#endif

#endif // #if defined (__C30__)


#define RXBUFSIZE   128
#define TXBUFSIZE   128
static char RxBuf[RXBUFSIZE];
static char TxBuf[TXBUFSIZE];
static volatile int TxWrtPtr;
static volatile int TxRdPtr;
static volatile int RxWrtPtr;
static volatile int RxRdPtr;

#if defined (__C30__)|| defined(__XC16__)

void __attribute__((__interrupt__,no_auto_psv)) _U1TXInterrupt(void)
{
	while(U1STAbits.UTXBF);
	IFS0bits.U1TXIF = 0;
	U1TXREG = TxBuf[TxRdPtr++];
	TxRdPtr &= (TXBUFSIZE-1);

	if(TxRdPtr == TxWrtPtr){
		IEC0bits.U1TXIE = 0; // disable TX interrupt
	}

}

void __attribute__((__interrupt__,no_auto_psv)) _U1RXInterrupt(void)
{
	char c;
	int  n;

	while(!U1STAbits.URXDA);
	c = U1RXREG;
	n = RxWrtPtr - RxRdPtr;
	if(n < 0) n+=RXBUFSIZE;

	if(n < (RXBUFSIZE-1)){ // return; // limit byte count is RXBUFSIZE-1
		RxBuf[RxWrtPtr++]=c;
		RxWrtPtr &= (RXBUFSIZE-1);
	}
        if(!U1STAbits.URXDA) IFS0bits.U1RXIF = 0;
}

#elif defined (__PIC32MX__)
// PIC32MX795F512L
void __attribute__((interrupt(ipl2soft), vector(_UART_1_VECTOR) )) _U1Interrupt(void)
{
	// error interrupt
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	if(IFS1bits.U1EIF){
		IFS1bits.U1EIF=0;
#else
	if(IFS0bits.U1EIF){
		IFS0bits.U1EIF=0;
#endif
		if(U1STAbits.OERR){
			U1STAbits.OERR=0;
		}
	}

	// RXinterrupt ?
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	if(IEC1bits.U1RXIE && IFS1bits.U1RXIF && U1STAbits.URXDA){
#else
	if(IEC0bits.U1RXIE && IFS0bits.U1RXIF && U1STAbits.URXDA){
#endif
		char c;
		int  n;

		c = U1RXREG;
		n = RxWrtPtr - RxRdPtr;
		if(n < 0) n+=RXBUFSIZE;

		if(n < (RXBUFSIZE-1)){ // return; // limit byte count is RXBUFSIZE-1
			RxBuf[RxWrtPtr++]=c;
			RxWrtPtr &= (RXBUFSIZE-1);
		}
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
		if(!U1STAbits.URXDA) IFS1bits.U1RXIF = 0;
#else
		if(!U1STAbits.URXDA) IFS0bits.U1RXIF = 0;
#endif
	}


	// TXinterrupt ?
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	if(IEC1bits.U1TXIE && IFS1bits.U1TXIF){ 
#else
	if(IEC0bits.U1TXIE && IFS0bits.U1TXIF){ 
#endif
		while(U1STAbits.UTXBF);
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
		IFS1bits.U1TXIF = 0;
#else
		IFS0bits.U1TXIF = 0;
#endif
		if(TxRdPtr == TxWrtPtr){
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
			IEC1bits.U1TXIE = 0; // disable TX interrupt
#else
			IEC0bits.U1TXIE = 0; // disable TX interrupt
#endif
			return; // no data in TxBuf
		}
		U1TXREG = TxBuf[TxRdPtr++];
		TxRdPtr &= (TXBUFSIZE-1);
	}
}

#else
    #error can not set interrupt function
#endif


/*******************************************************************************
Function: UART1GetBaudError()

Precondition:
    None.

Overview:
    This routine checks the UART baud rate error percentage and returns it.

Input: None.

Output: Returns the baud rate error in percent.

*******************************************************************************/
char UART1GetBaudError()
{
    unsigned int errorPercent = 0;

	errorPercent = ((BAUD_ERROR*100+BAUDRATE1/2)/BAUDRATE1);
    return (char)errorPercent;
}


/*******************************************************************************
Function: UART1GetChar()

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This routine waits for a byte to be received.  It then returns that byte.

Input: None.

Output: Byte received.

*******************************************************************************/
#if 0
char UART1GetChar()
{
    char Temp;

    while(IFS0bits.U1RXIF == 0);

    Temp = U1RXREG;
    IFS0bits.U1RXIF = 0;
    return Temp;
}
#endif

char UART1GetChar()
{
	char c;

	while(RxRdPtr == RxWrtPtr); // wait for received data
	c = RxBuf[RxRdPtr];
#if defined (__C30__)|| defined(__XC16__)
	IEC0bits.U1RXIE=0;
#elif defined (__PIC32MX__)
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1RXIE=0;
#else
	IEC0bits.U1RXIE=0;
#endif
#endif

	RxRdPtr++;
	RxRdPtr &= (RXBUFSIZE-1);

#if defined (__C30__)|| defined(__XC16__)
	IEC0bits.U1RXIE=1;
#elif defined (__PIC32MX__)
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1RXIE=1;
#else
	IEC0bits.U1RXIE=1;
#endif
#endif

	return c;
}



char *UART1GetString(char *buf, int size)
{
	int c, len;

	len=0;

	while(len<(size-1)){
		c = UART1GetChar();
        UART1PutChar(c);
		if(c == '\r'){
            UART1PutChar('\n');
			buf[len] = 0;
			return buf;
		}
		buf[len++]=c;
	}
	buf[len]=0;
	return buf;

}


void UART1GetData(unsigned char *buf, int len)
{
	while(len--){
		*buf++ = UART1GetChar();
	}
}



/*******************************************************************************
Function: UART1Init()

Precondition: None.

Overview:
    This routine sets up the UART1 module.

Input: None.

Output: None.

Notes:
    Allow the peripheral to set the I/O pin directions.  If we set the TRIS
    bits manually, then when we disable the UART, the shape of the stop bit
    changes, and some terminal programs have problems.
*******************************************************************************/
void UART1Init()
{
    U1BRG = BAUDRATEREG1;
    U1MODE = 0;
    U1MODEbits.BRGH = BRGH1;
    U1STA = 0;
    U1MODEbits.UARTEN = 1;

#if defined(__C30__)|| defined(__XC16__)
    U1STAbits.UTXISEL1 = 0;
    U1STAbits.UTXISEL0 = 1;
    U1STAbits.URXISEL = 0;
    U1STAbits.UTXEN = 1;
    IFS0bits.U1RXIF = 0;
    IFS0bits.U1TXIF = 0;
#elif defined (__PIC32MX__)
    U1STAbits.UTXISEL = 0;
    U1STAbits.URXISEL = 0;
    U1STAbits.UTXEN = 1;
    U1STAbits.URXEN = 1;
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
    IFS1bits.U1RXIF = 0;
    IFS1bits.U1TXIF = 0;
    IFS1bits.U1EIF = 0;
#else
    IFS0bits.U1RXIF = 0;
    IFS0bits.U1TXIF = 0;
    IFS0bits.U1EIF = 0;
#endif
#else
	#error can not clear interrupt flag
#endif

    RxRdPtr = RxWrtPtr = 0;
    TxRdPtr = TxWrtPtr = 0;

#if defined(__C30__)|| defined(__XC16__)
    IPC2bits.U1RXIP = 7; // UART RX interrupt priority is 7
    IPC3bits.U1TXIP = 7; // UART TX interrupt priority is 7
    IEC0bits.U1RXIE = 1; // Enable RX interrupt
//  IEC0bits.U1TXIE = 1; // do not enable this timing,
#elif defined (__PIC32MX__)
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
    IPC8bits.U1IP = 2;   // UART 1 interrupt priority is 2
    IEC1bits.U1RXIE = 1; // Enable RX interrupt
//  IEC1bits.U1TXIE = 1; // do not enable this timing
    IEC1bits.U1EIE = 1;  // Enable ERR interrupt
#else
    IPC6bits.U1IP = 2;   // UART 1 interrupt priority is 2
    IEC0bits.U1RXIE = 1; // Enable RX interrupt
//  IEC0bits.U1TXIE = 1; // do not enable this timing
    IEC0bits.U1EIE = 1;  // Enable ERR interrupt
#endif
#else
	#error can not define priority
#endif
}


void UART1RxBufClr(void)
{
//    while(UART1IsPressed()) UART1GetChar();

#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1RXIE=0;
#else
	IEC0bits.U1RXIE=0;
#endif
	RxRdPtr = RxWrtPtr = 0;	
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1RXIE=1;
#else
	IEC0bits.U1RXIE=1;
#endif

}


/*******************************************************************************
Function: UART1IsPressed()

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This routine checks to see if there is a new byte in UART reception buffer.

Input: None.

Output:
    0 : No new data received.
    1 : Data is in the receive buffer

*******************************************************************************/
#if 0
char UART1IsPressed()
{
    if(IFS0bits.U1RXIF == 1)
        return 1;
    return 0;
}
#endif

char UART1IsPressed()
{
    if(RxRdPtr == RxWrtPtr) return 0;
    return 1;
}


/*******************************************************************************
Function: UART1PrintString( char *str )

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This function prints a string of characters to the UART.

Input: Pointer to a null terminated character string.

Output: None.

*******************************************************************************/
void UART1PrintString( char *str )
{
    unsigned char c;

    while( (c = *str++) )
        UART1PutChar(c);
}

void UART1PutData(unsigned char *data, int len)
{
	while(len--){
		UART1PutChar(*data++);
	}
}

/*******************************************************************************
Function: UART1PutChar( char ch )

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This routine writes a character to the transmit FIFO, and then waits for the
    transmit FIFO to be empty.

Input: Byte to be sent.

Output: None.

*******************************************************************************/
#if 0
void UART1PutChar( char ch )
{
	while(U1STAbits.UTXBF == 1);	// add suwa

    U1TXREG = ch;
    #if !defined(__PIC32MX__)
        Nop();
        Nop();
    #endif
    while(U1STAbits.TRMT == 0);
}
#endif

void UART1PutChar(char ch)
{
	int n;

	do {
		n = TxWrtPtr - TxRdPtr;
		if(n < 0) n += TXBUFSIZE;
	}while(n >=(TXBUFSIZE-1)); // buffer full

#if defined(__C30__) || defined(__XC16__)
	TxBuf[TxWrtPtr] = ch;
	IEC0bits.U1TXIE = 0; // lock interrupt
	TxWrtPtr++;
	TxWrtPtr &= (TXBUFSIZE-1);
	IEC0bits.U1TXIE = 1; // unlock interrupt
#elif defined(__PIC32MX__)
	TxBuf[TxWrtPtr] = ch;
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1TXIE = 0; // lock interrupt
#else
	IEC0bits.U1TXIE = 0; // lock interrupt
#endif
	TxWrtPtr++;
	TxWrtPtr &= (TXBUFSIZE-1);
#if defined(__32MX220F032B__) || defined (__32MX210F016B__)
	IEC1bits.U1TXIE = 1; // unlock interrupt
#else
	IEC0bits.U1TXIE = 1; // unlock interrupt
#endif
#endif

}


/*******************************************************************************
Function: UART1PutDec(unsigned char dec)

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This function converts decimal data into a string and outputs it to UART.

Input: Binary data.

Output: None.

*******************************************************************************/
void  UART1PutDec(unsigned char dec)
{
    unsigned char res;
    unsigned char printed_already = 0;

    res = dec;

    if (res/100)
    {
        UART1PutChar( res/100 + '0' );
        printed_already = 1;
    }
    res = res - (res/100)*100;

    if ((res/10) || (printed_already == 1))
    {
        UART1PutChar( res/10 + '0' );
    }
    res = res - (res/10)*10;

    UART1PutChar( res + '0' );
}

/*******************************************************************************
Function: UART1PutHex

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This function converts hex data into a string and outputs it to UART.

Input: Binary data.

Output: None.

*******************************************************************************/

static const unsigned char CharacterArray[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

void UART1PutHex( int toPrint )
{
    int printVar;

    printVar = toPrint;
    toPrint = (toPrint>>4) & 0x0F;
    UART1PutChar( CharacterArray[toPrint] );

    toPrint = printVar & 0x0F;
    UART1PutChar( CharacterArray[toPrint] );

    return;
}

/*******************************************************************************
Function: UART1PutHexWord(unsigned int toPrint)

Precondition:
    UART1Init must be called prior to calling this routine.

Overview:
    This function converts hex data into a string and outputs it to UART.

Input: Binary data.

Output: None.

*******************************************************************************/
#if defined( __C30__ ) || defined ( __XC16__ ) || defined( __PIC32MX__ )
void UART1PutHexWord( unsigned int toPrint )
{
    unsigned int printVar;

    printVar = (toPrint>>12) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>8) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>4) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = toPrint & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    return;
}

void UART1PutHexDWord( unsigned long toPrint )
{
    unsigned long printVar;

    printVar = (toPrint>>28) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>24) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>20) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>16) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>12) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>8) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = (toPrint>>4) & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    printVar = toPrint & 0x0F;
    UART1PutChar( CharacterArray[printVar] );

    return;
}

#endif

/*********************************************************************
Function: char UART1Char2Hex(char ch)

PreCondition: none

Input: ASCII to be converted

Output: number

Side Effects: none

Overview: converts ASCII coded digit into number

Note: none

********************************************************************/
char UART1Char2Hex(char ch){
  // Wrong char
  if(ch > 102)
    return 0;  
  
  // From a to f     
  if(ch > 96)
    return (ch-87);
    
  // Wrong char    
  if(ch > 70)
    return 0;
  
  // From A to F    
  if(ch > 64)
    return (ch-55);

  // Wrong char  
  if(ch > 57)
    return 0;
  
  // From 0 - 9
  if(ch > 47)
    return(ch-48);
  else
  // Wrong char
    return 0;
}

/*********************************************************************
Function: char UART1Hex2Char(char hex)

PreCondition: none

Input: number

Output: ASCII code

Side Effects: none

Overview: converts low nibble into ASCII coded digit

Note: none

********************************************************************/
char UART1Hex2Char(char hex){
char h;
  h = hex&0x0f;
  // From 0xa to 0xf
  if(h>9)
    return (h+55);
  else
    return (h+48);
}

/*********************************************************************
Function: void UART1ClrError(void)

PreCondition: none

Input: none

Output: character received

Side Effects: none

Overview: wait for character

Note: none

********************************************************************/
void UART1ClrError(void){
    // Clear error flag
    if(U1STAbits.OERR)
		U1STAbits.OERR = 0;
}

